//
// This file is automatically generated by generate_auto_job.
// Edits will be automatically overwritten if the build is
// run in maintainer mode.
//
// clang-format off
//
void setupInputFile();
void setupPassword();
void setupEmpty();
void setupOutputFile();
void setupReplaceInput();
void beginEncrypt(JSON);
void endEncrypt();
void setupEncryptUserPassword();
void setupEncryptOwnerPassword();
void beginEncrypt40bit(JSON);
void endEncrypt40bit();
void beginEncrypt128bit(JSON);
void endEncrypt128bit();
void beginEncrypt256bit(JSON);
void endEncrypt256bit();
void beginJsonKeyArray(JSON);
void endJsonKeyArray();
void beginJsonObjectArray(JSON);
void endJsonObjectArray();
void beginGlobal(JSON);
void endGlobal();
void beginAddAttachmentArray(JSON);
void endAddAttachmentArray();
void beginAddAttachment(JSON);
void endAddAttachment();
void setupAddAttachmentFile();
void beginRemoveAttachmentArray(JSON);
void endRemoveAttachmentArray();
void beginCopyAttachmentsFromArray(JSON);
void endCopyAttachmentsFromArray();
void beginCopyAttachmentsFrom(JSON);
void endCopyAttachmentsFrom();
void setupCopyAttachmentsFromFile();
void setupCopyAttachmentsFromPassword();
void beginPagesArray(JSON);
void endPagesArray();
void beginPages(JSON);
void endPages();
void setupPagesFile();
void setupPagesPassword();
void beginRotateArray(JSON);
void endRotateArray();
void beginSetPageLabelsArray(JSON);
void endSetPageLabelsArray();
void setupSetPageLabels();
void beginOverlayArray(JSON);
void endOverlayArray();
void beginOverlay(JSON);
void endOverlay();
void setupOverlayFile();
void setupOverlayPassword();
void beginUnderlayArray(JSON);
void endUnderlayArray();
void beginUnderlay(JSON);
void endUnderlay();
void setupUnderlayFile();
void setupUnderlayPassword();
